import typing
from functools import reduce
from abc import ABC, abstractmethod
from .._core import Message, render
from ..utils import Renderable


def get_or_spawn(state: typing.Dict, child: str) -> typing.Dict:
    """Get a child or spawn it if it does not exist

    Args:
        state (typing.Dict): The state
        child (str): The name of the child

    Returns:
        typing.Dict: The dictionary for the child
    """
    if child not in state:
        state[child] = {}
    return state[child]


def get_or_set(state: typing.Dict, key: str, val: typing.Any) -> typing.Any:
    """Use to get or set a value in a state dict

    Args:
        state (typing.Dict): The state dict to update
        key (str): The key to update
        val (typing.Any): The value to add

    Returns:
        typing.Any: 
    """
    if key not in state:
        state[key] = val
    return state[key]


class SharedBase(Renderable, ABC):
    """Allows for shared data between tasks
    """

    @abstractmethod
    def register(self, callback) -> bool:
        """Register a callback to call on data updates

        Args:
            callback (function): The callback to register

        Returns:
            bool: True the callback was registered, False if already registered
        """
        pass

    @abstractmethod
    def unregister(self, callback) -> bool:
        """Unregister a callback to call on data updates

        Args:
            callback (function): The callback to unregister

        Returns:
            bool: True if the callback was removed, False if callback was not registered
        """
        pass

    @property
    @abstractmethod
    def data(self) -> typing.Any:
        """Get the data shared

        Returns:
            typing.Any: The shared data
        """
        return self._data

    @data.setter
    @abstractmethod
    def data(self, data) -> typing.Any:
        """Update the data

        Args:
            data: The data to share

        Returns:
            typing.Any: The value for the data
        """
        pass

    @abstractmethod
    def get(self) -> typing.Any:
        pass
    
    @abstractmethod
    def set(self, value) -> typing.Any:
        pass

    def render(self) -> str:
        return render(self.get())


class Shared(SharedBase):
    """Allows for shared data between tasks
    """
    def __init__(
        self, data: typing.Any=None, 
        default: typing.Any=None,
        callbacks: typing.List[typing.Callable[[typing.Any], None]]=None
    ):
        """Create shard data

        Args:
            data (typing.Any, optional): The data. Defaults to None.
            default (typing.Any, optional): The default value of the shared. When resetting, will go back to this. Defaults to None.
            callbacks (typing.List[typing.Callable[[typing.Any], None]], optional): The callbacks to call on update. Defaults to None.
        """
        self._default = default
        self._data = data if data is not None else default
        self._callbacks = callbacks or []

    def register(self, callback) -> bool:
        """Register a callback to call on data updates

        Args:
            callback (function): The callback to register

        Returns:
            bool: True the callback was registered, False if already registered
        """
        if callback in self._callbacks:
            return False
        
        self._callbacks.append(callback)
        return True

    def unregister(self, callback) -> bool:
        """Unregister a callback to call on data updates

        Args:
            callback (function): The callback to unregister

        Returns:
            bool: True if the callback was removed, False if callback was not registered
        """

        if callback not in self._callbacks:
            return False
        
        self._callbacks.remove(callback)
        return True

    @property
    def data(self) -> typing.Any:
        """Get the data shared

        Returns:
            typing.Any: The shared data
        """
        return self._data

    @data.setter
    def data(self, data) -> typing.Any:
        """Update the data

        Args:
            data: The data to share

        Returns:
            typing.Any: The value for the data
        """
        self._data = data
        for callback in self._callbacks:
            callback(data)
        return data

    def get(self) -> typing.Any:
        """Get the value

        Returns:
            typing.Any: 
        """
        return self._data
    
    def set(self, value) -> typing.Any:
        """Set the shared value

        Args:
            value : Set the shared value

        Returns:
            typing.Any: The value the shared value was set to
        """
        self.data = value
        return value
    
    def reset(self):
        """Reset the shared value back to the default value
        """
        self.data = self._default


class Buffer(object):
    """Create a buffer to add data to
    """

    def __init__(self) -> None:
        """Create a buffer
        """
        self._buffer = []
        self._opened = True
        
    def add(self, *data):
        """Add data to the buffer

        Raises:
            RuntimeError: if the buffer is closed
        """
        
        if not self._opened:
            raise RuntimeError(
                'The buffer is currently closed so cannot add data to it.'
            )

        self._buffer.extend(data)

    def open(self):
        """Open the buffer
        """
        self._opened = True

    def close(self):
        """Close the buffer
        """
        self._opened = False

    def is_open(self) -> bool:
        """Get whether the buffer is open

        Returns:
            bool: Whether the buffer is open
        """
        return self._opened
        
    def it(self) -> 'BufferIter':
        return BufferIter(self)

    def __getitem__(self, idx) -> typing.Union[typing.Any, typing.List]:
        """Get a value from the buffer

        Args:
            idx: The index to to retrieve

        Returns:
            typing.Union[typing.Any, typing.List]: The retrieved data
        """
        if isinstance(idx, slice):
            return self._buffer[idx]
        if isinstance(idx, typing.Iterable):
            return [self._buffer[i] for i in idx]
        return self._buffer[idx]
        
    def __len__(self) -> int:
        """Get the length of the buffer

        Returns:
            int: the length
        """
        return len(self._buffer)
    
    def get(self):
        return [*self._buffer]


class BufferIter(object):
    """An iterator to retrieve values from a Buffer
    """

    def __init__(self, buffer: Buffer) -> None:
        """Create a buffer iterator

        Args:
            buffer (Buffer): The buffer to iterate over
        """
        self._buffer = buffer
        self._i = 0

    def start(self):
        """Whether the iterator is at the start of the buffer
        """
        return self._i == 0

    def end(self) -> bool:
        """Get whether the buffer is at an end

        Returns:
            whether it is the end of the buffer iter
        """
        return self._i >= (len(self._buffer) - 1)
    
    def is_open(self) -> bool:
        """Get if the buffer is open

        Returns:
            bool: Whether the buffer is open
        """
        return self._buffer.is_open()

    def read(self) -> typing.Any:
        """Read the next value in the buffer

        Raises:
            StopIteration: If the buffer is at an end

        Returns:
            typing.Any: The next value in the buffer
        """
        if self._i < (len(self._buffer) - 1):
            self._i += 1
            return self._buffer[self._i - 1]
        raise StopIteration('Reached the end of the buffer')

    def read_all(self) -> typing.List:
        """_summary_

        Returns:
            typing.List: _description_
        """
        
        if self._i < len(self._buffer):
            i = self._i
            self._i = len(self._buffer)
            return self._buffer[i:]
            
        return []

    def read_reduce(self, f, init_val=None) -> typing.Any:

        data = self.read_all()
        return reduce(
            f, data, init_val
        )

    def read_map(self, f) -> typing.List:

        data = self.read_all()
        return map(
            f, data
        )


class Context(dict):
    """Use to store state
    """

    def get_or_set(self, key, value):
        """Get or set a value in the context

        Args:
            key : The key for the value to set
            value : The value to set if not already set

        Returns:
            Any: the value specified by key
        """
        if key not in self:
            self[key] = value
            return value
        
        return self[key]


class ContextStorage(object):
    """Use to manage context storage such as spawning new
    contexts.
    """

    def __init__(self):
        """Create the context storage
        """
        object.__setattr__(self, '_data', {})
        self._data: typing.Dict

    @property
    def contexts(self) -> typing.Dict[str, Context]:
        """Retrieve all contexts

        Returns:
            typing.Dict[str, Context]: All contexts
        """
        return {**self._data}
    
    def clear(self):
        """Clear the entire context storage
        """
        self._data = {}

    def reset(self):
        """Clear the entire context storage
        """
        self.clear()
    
    def remove(self, key):
        """Remove a context

        Args:
            key: The name of the context to remove
        """
        del self._data[key]
    
    def add(self, key) -> 'Context':
        """Add context to the storage

        Args:
            key : The name of the context 

        Returns:
            Context: The spawned context
        """
        if key not in self._data:
            d = Context()
            self._data[key] = d
        
        return self._data[key]

    def __getattr__(self, key) -> 'Context':
        """Retrieve the context

        Args:
            key Retrieve the context

        Returns:
            Context: The created context
        """
        if key not in self._data:
            d = Context()
            self._data[key] = d
        
        return self._data[key]


class Blackboard(object):
    """A blackboard is for sharing information
    across tasks
    """

    def __init__(self):
        """The data for the blackboard
        """
        object.__setattr__('_data', {})
        object.__setattr__('_callbacks', {})
        self._data: typing.Dict
        self._callbacks: typing.Dict[str, typing.List[typing.Callable[[typing.Any]]]]

    def r(self, key) -> 'Retriever':
        """Get a retriever for the blackboard

        Args:
            key: The name of the key for the retriever

        Returns:
            Retriever: The retriever for the blackboard
        """
        return Retriever(self, key)

    def register(self, key, callback) -> bool:
        """Register a callback to call on data updates

        Args:
            callback (function): The callback to register

        Returns:
            bool: True the callback was registered, False if already registered
        """
        if key not in self._callbacks:
            self._callbacks[key] = [callable]
            return True

        if callback in self._callbacks[key]:
            return False
        
        self._callbacks[key].append(callback)
        return True

    def unregister(self, key, callback) -> bool:
        """Unregister a callback to call on data updates

        Args:
            callback (function): The callback to unregister

        Returns:
            bool: True if the callback was removed, False if callback was not registered
        """
        if key not in self._callbacks:
            return False

        if callback not in self._callbacks[key]:
            return False
        
        self._callbacks[key].remove(callback)
        return True

    def __getitem__(self, key) -> typing.Any:
        """Get an item from the blackboard

        Args:
            key: The key to retrieve for

        Returns:
            typing.Any: The value for the key
        """
        return self._data[key]

    def __setitem__(self, key, val) -> typing.Any:
        """Set an item in the blackboard

        Args:
            key: The name of the item to set
            val: The value to set to

        Returns:
            typing.Any: The name of the 
        """
        self._data[key] = val
        return val

    def __getattr__(self, key):

        if key not in self._data:
            d = Context()
            self._data[key] = d
        
        return self._data[key]


class Retriever(SharedBase):
    """Use to retrieve data and set data in the blackboard
    """

    def __init__(self, blackboard: Blackboard, key: str):
        """Create a retriever to retrieve data from the blackboard

        Args:
            blackboard (Blackboard): The blackboard to retrieve from
            key (str): The key to retrieve from
        """
        self._blackboard = blackboard
        self._key = key

    def get(self) -> typing.Any:
        """Get a value from the blackboard

        Returns:
            typing.Any: The value to get
        """
        return self._blackboard[self._key]
    
    def set(self, val) -> typing.Any:
        """Set the value the retriever points to

        Args:
            val: The value to set to

        Returns:
            typing.Any: The value to set
        """
        self._blackboard[self._key] = val
        return val

    def register(self, callback) -> bool:
        """Register a callback to call on data updates

        Args:
            callback (function): The callback to register

        Returns:
            bool: True the callback was registered, False if already registered
        """
        self._blackboard.register(self._key, callback)

    def unregister(self, callback) -> bool:
        """Unregister a callback to call on data updates

        Args:
            callback (function): The callback to unregister

        Returns:
            bool: True if the callback was removed, False if callback was not registered
        """
        self._blackboard.unregister(self._key, callback)

    @property
    def data(self) -> typing.Any:
        """Get the data shared

        Returns:
            typing.Any: The shared data
        """
        return self._blackboard[self._key]

    @data.setter
    def data(self, data) -> typing.Any:
        """Update the data

        Args:
            data: The data to share

        Returns:
            typing.Any: The value for the data
        """
        self._blackboard[self._key] = data
        return data


class ContextSpawner(object):
    """Use to Spawn contexts for your behaviors
    """

    def __init__(self, manager: ContextStorage, base_name: str):
        """Create a context Spawner specifying the Storage and the
        base name to use in spawning

        Args:
            manager (): The manager to use
            base_name (str): The base name of all contexts to spawn
        """
        object.__setattr__('manager', manager)
        object.__setattr__('base_name', base_name)

    def __getitem__(self, name) -> Context:
        """Get item specified by name

        Args:
            name: The context to get

        Returns:
            Context: The context
        """

        name = f'{self.base_name}_{name}'
        return self.manager.add(name)

    def __getattr__(self, name: int) -> Context:
        """Get item specified by name

        Args:
            name: The context to get

        Returns:
            Context: The context
        """
        name = f'{self.base_name}_{name}'
        return self.manager.add(name)


class Comm(object):
    """Use to have communication between two components 
    (two agents etc)
    """

    def __init__(self):
        """
        """
        self._processing = []

    def post(self, message: Message, callback: typing.Callable[[], Message]=None):
        """Post a message to be received

        Args:
            message (Message): 
            callback (optional): . Defaults to None.
        """
        self._processing.append((message, callback))

    def get(self) -> Message:
        """Get the current message to process

        Raises:
            RuntimeError: 

        Returns:
            Message: The message to process
        """
        if self.empty():
            raise RuntimeError('No items to process')
        return self._processing[0][0]
    
    def empty(self) -> bool:
        """If nothing is being processed

        Returns:
            bool: Whether nothing is waiting for processing
        """
        return len(self._processing) == 0

    def respond(self, with_message: Message):

        if self.empty():
            raise RuntimeError(
                'Cannot respond with a message '
            )
        
        callback = self._processing[0][1]
        if callback is not None:
            callback(with_message)
        self._processing.pop(0)



from dachi.act._tasks import Task
from . import _tasks as behavior
import typing


class build_sango(object):
    """Base context method to build a behavior tree
    """

    def __init__(self) -> None:
        """Create the behavior tree. This is the root node
        """
        super().__init__()
        self._sango = behavior.Root()

    def __enter__(self):
        return self._sango
    
    def __exit__(self, exc_type, exc_value, traceback):

        if exc_type is not None:
            raise


class build_composite(object):
    """Base context method to build a task decorator
    """

    def __init__(self, child: behavior.Task, parent: behavior.Task=None) -> None:
        """Create a composite node that uses a list to store all of the subtasks

        Args:
            name (str): The name of the "Sango"
        """
        super().__init__()
        self._child = child
        self._parent = parent
        
        if parent is None:
            pass
        elif isinstance(parent, behavior.Root):
            parent.root = child
        elif isinstance(parent, behavior.Serial):
            parent.add(child)
        elif isinstance(parent, behavior.Parallel):
            parent.add(child)

    @property
    def task(self) -> behavior.Task:
        """
        Returns:
            behavior.Task: The composite task
        """
        return self._child

    @property
    def parent(self) -> behavior.Task:
        """
        Returns:
            behavior.Task: The parent task of the composite task
        """
        return self._parent

    def __enter__(self) -> behavior.Serial:
        return self._child
    
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            raise


class build_decorate(object):
    """Base context method to build a task decorator
    """

    def __init__(
        self, decorate: typing.Type[behavior.Decorator], 
        decorated: typing.Union['build_decorate', build_composite], parent: behavior.Task=None,
        **kwargs
    ) -> None:
        """Decorate the task

        """
        super().__init__()
        assert decorated.parent is None
        self._decorated = decorate(decorated.task, **kwargs)

        self._parent = parent
        
        if parent is None:
            pass
        elif isinstance(parent, behavior.Root):
            parent.root = self._decorated
        elif isinstance(parent, behavior.Serial):
            parent.add(self._decorated)
        elif isinstance(parent, behavior.Parallel):
            parent.add(self._decorated)

    @property
    def task(self) -> behavior.Task:
        """
        Returns:
            behavior.Task: The decorated task
        """
        return self._decorated

    @property
    def parent(self) -> behavior.Task:
        """
        Returns:
            behavior.Task: The parent of the decorated tak
        """
        return self._parent

    def __enter__(self):
        return self._decorated
    
    def __exit__(self, exc_type, exc_value, traceback):

        if exc_type is not None:
            raise


class build_sequence(build_composite):
    """Context method to build a sequence task
    """

    def __init__(self, parent: behavior.Task=None) -> None:
        """Create a Sequence task

        Args:
            parent (behavior.Task, optional): The parent of the sequence. Defaults to None.
        """
        super().__init__(behavior.Sequence([]), parent)


class build_select(build_composite):
    """Context method to build a Selector task
    """

    def __init__(self, parent: behavior.Task=None) -> None:
        """Create a selector task

        Args:
            parent (behavior.Task, optional): The parent of the selector. Defaults to None.
        """
        super().__init__(behavior.Selector([]), parent)


build_fallback = build_select


class build_parallel(build_composite):
    """Context manager to build an Parallel task
    """

    def __init__(
        self, parent: behavior.Task=None, 
        fails_on: int=1, succeeds_on: int=None, success_priority: bool=False
    ):
        """Execute two tasks in parallel

        Args:
            parent (typing.Union[behavior.Sango, CompositeBuilder], optional): The parent biulder. Defaults to None.
            fails_on (int, optional): The number of tasks that must fail for this to fail. Defaults to 1.
            succeeds_on (int, optional): The number of tasks to succeed for the parallel task to succeed. Defaults to None.
            success_priority (bool, optional): Whether it will return as soon as it succeeds. Defaults to False.
        """
        parallel = behavior.Parallel(
            [], fails_on=fails_on, succeeds_on=succeeds_on, 
            success_priority=success_priority)
        super().__init__(
            parallel, parent
        )


class build_not(build_decorate):
    """Context manager to build an Not decorated task
    """

    def __init__(self, decorated: build_composite, parent: Task = None) -> None:
        """Invert the output of a composite task

        Args:
            decorated (composite): The task to decorate
            parent (Task, optional): The parent task. Defaults to None.
        """
        super().__init__(behavior.Not, decorated, parent)


class build_unless(build_decorate):
    """Context manager to build an Unless decorated task
    """

    def __init__(self, decorated, parent: Task = None, target_status: behavior.TaskStatus=behavior.TaskStatus.FAILURE) -> None:
        """Loop over the subtask while it 'succeeds'

        Args:
            decorated: The task to loop over
            parent (Task, optional): The parent task. Defaults to None.
        """
        super().__init__(behavior.Unless, decorated, parent, target_status=target_status)


class build_until(build_decorate):
    """Context manager to build an Until decorated task
    """

    def __init__(self, decorated: build_composite, parent: Task = None, target_status: behavior.TaskStatus=behavior.TaskStatus.SUCCESS) -> None:
        """Loop over the subtask until it 'succeeds'

        Args:
            decorated: The task to loop over
            parent (Task, optional): The parent task. Defaults to None.
        """
        super().__init__(behavior.Until, decorated, parent, target_status=target_status)



# 1st party
import typing
import string
import re
import pydantic
from enum import Enum


class _PartialFormatter(string.Formatter):

    def __init__(self):
        super().__init__()

    def format(self, format_string, *args, required: bool=True, **kwargs):
        """Format the string

        Args:
            format_string : The string to format
            required (bool, optional): Whether the key is required. Defaults to True.

        Returns:
            str: the formatted string
        """
        if args and kwargs:
            raise ValueError("Cannot mix positional and keyword arguments")

        if kwargs and required:
            difference = set(kwargs.keys()).difference(set(get_str_variables(format_string)))
            if difference:
                raise ValueError(f'Variables specified that are not in the string {difference}')
        self.args = args
        self.kwargs = kwargs
        return super().format(format_string)

    def get_value(self, key, args, kwargs):
        if isinstance(key, str):
            return self.kwargs.get(key, '{' + key + '}')
        if isinstance(key, int):
            return self.args[key] if key < len(self.args) else '{' + str(key) + '}'
        return super().get_value(key, args, kwargs)

    def __call__(self, format_string, *args, **kwargs):
        return self.format(format_string, *args, **kwargs)


def get_str_variables(format_string: str) -> typing.List[str]:
    """Get the variables in a string to format

    Args:
        format_string (str): The string to get variables for 

    Raises:
        ValueError: If the string has both positional and named
        variables

    Returns:
        typing.List[str]: The list of variables
    """
    has_positional = re.search(r'\{\d*\}', format_string)
    has_named = re.search(r'\{[a-zA-Z_]\w*\}', format_string)
    
    if has_positional and has_named:
        raise ValueError("Cannot mix positional and named variables")

    # Extract variables
    if has_positional:
        variables = [int(var) if var.isdigit() else None for var in re.findall(r'\{(\d*)\}', format_string)]
        if None in variables:
            variables = list(range(len(variables)))
    else:
        variables = re.findall(r'\{([a-zA-Z_]\w*)\}', format_string)
    
    return variables


str_formatter = _PartialFormatter() # Use to format a string with keys and values, does not require all keys specified


def escape_curly_braces(value: typing.Any, render: bool=False) -> str:
    """Escape curly braces for dictionary-like structures."""

    if isinstance(value, str):
        result = f'"{value}"'
        return result
    if isinstance(value, typing.Dict):
        items = ', '.join(f'"{k}": {escape_curly_braces(v)}' for k, v in value.items())
        return f"{{{{{items}}}}}"
    if isinstance(value, typing.List):
        return '[{}]'.format(', '.join(escape_curly_braces(v) for v in value))
    if render:
        return render(value)
    return str(value)


def unescape_curly_braces(value: typing.Any) -> str:
    """Invert the escaping of curly braces."""
    if isinstance(value, str):
        return value.replace('{{', '{').replace('}}', '}')
    return value


primitives = (bool, str, int, float, type(None))
"""a list of primitive types"""


def is_primitive(obj) -> bool:
    """Utility to check if a value is a primitive

    Args:
        obj: Value to check

    Returns:
        bool: If it is a "primitive"
    """
    return type(obj) in primitives


def generic_class(t: typing.TypeVar, idx: int=0) -> typing.Type:
    """Gets the generic type for a class assuming that it only has 
    one.

    Args:
        t (typing.TypeVar): The class to get the generic type for
        idx (int, optional): . Defaults to 0.

    Returns:
        typing.Type: the type specified by the generic class
    """
    return t.__orig_class__.__args__[idx]


class _Types(Enum):

    UNDEFINED = 'UNDEFINED'
    WAITING = 'WAITING'


UNDEFINED = _Types.UNDEFINED # Constant for UNDEFINED. usage: value is UNDEFINED
WAITING = _Types.WAITING # Constant for WAITING when streaming. usage: value is WAITING


def is_nested_model(
    pydantic_model_cls: typing.Type[pydantic.BaseModel]
) -> bool:
    """Helper function to check if it is a nested model

    Args:
        pydantic_model_cls (typing.Type[pydantic.BaseModel]): The class to check if it is a nested model

    Returns:
        bool: If it is a nested model
    """
    for field in pydantic_model_cls.model_fields.values():
        
        if isinstance(field.annotation, type) and issubclass(field.annotation, pydantic.BaseModel):
            return True
    return False


def is_undefined(val) -> bool:
    """Returns true if the vlaue is undefined

    Args:
        val : The value to check

    Returns:
        bool: Whether the value is undefined or not
    """
    return val == UNDEFINED or val == WAITING


def get_member(obj, loc: str):
    """Get a member from an object recursively

    Args:
        obj : the object
        loc (str): the location as a string, use '.' to indicate sub objects

    Returns:
        Any: The member
    """
    locs = loc.split('.')
    for loc in locs:
        obj = getattr(obj, loc)
    return obj



# 1st party
import typing
from abc import ABC, abstractmethod
from typing import Self
import typing

import asyncio
from ._core import (
    Instruct, Cue, Reader, 
    render, NullRead, Module
)
from dataclasses import dataclass
from ._core import Renderable

# 3rd party
import pydantic


@dataclass
class AIResponse(object):
    """The AI response stores the response from the API
    And any processing done on it.
    """

    message: 'Message' # The message from the AI
    source: typing.Dict # The response from the API
    val: typing.Any = None # the result of processing the message

    def clone(self) -> Self:

        return AIResponse(
            message=self.message,
            source=self.source,
            val=self.val
        )
    
    def __iter__(self) -> typing.Iterator:
        """Unpack the values in the AIResponse

        Yields:
            Iterator[typing.Iterator]: Each value in the response
        """
        yield self.source
        yield self.message
        yield self.val


class AIPrompt(pydantic.BaseModel, ABC):
    """Base class for prompts to send to the AI
    """

    @abstractmethod
    def prompt(self, model: 'AIModel'):
        """Send the prompt to the AI Model

        Args:
            model (AIModel): The model to query
        """
        pass

    @abstractmethod
    def instruct(self, cue: 'Cue', model: 'AIModel'):
        """Send an instruction to the model

        Args:
            cue (Cue): The cue to use
            model (AIModel): The model to use
        """
        pass

    @abstractmethod
    def reader(self) -> 'Reader':
        """Get the reader from the AIPrompt to process the response

        Returns:
            Reader: The reader used to compute val in the response
        """
        pass

    @abstractmethod
    def process_response(self, response: 'AIResponse') -> 'AIResponse':
        """Process the response of the Model

        Args:
            response (AIResponse): The response to process

        Returns:
            AIResponse: The updated response
        """
        pass

    @abstractmethod
    def clone(self) -> typing.Self:
        """Do a shallow clone of the message object

        Returns:
            typing.Self: The cloned object
        """
        pass

    @abstractmethod
    def aslist(self) -> typing.List['Message']:
        """Return the AIPrompt as a list of messages
        """
        pass

    def __iter__(self) -> typing.Iterator['Message']:
        """Iterate over each message in the prompt

        Yields:
            Message: Each message
        """
        for message in self.aslist():
            yield message


class Message(pydantic.BaseModel, Renderable):
    """A prompt that consists of a single message to send the AI
    """

    source: str # The source of the messsage (user, assistant etc)
    data: typing.Dict[str, typing.Any] # The contents of the message

    def __getitem__(self, key: str) -> typing.Any:
        """Get an item from the message

        Args:
            key (str): The key to retrieve for

        Raises:
            KeyError: 

        """
        if hasattr(self, key):
            return getattr(self, key)
        if key in self.data:
            return self.data[key]
        raise KeyError(f'{key}')

    def __setitem__(self, key: str, value: typing.Any):
        """Set an item in the message

        Args:
            key (str): The key to set for
            value (typing.Any): The value to set

        Raises:
            KeyError: An error
        """
        if hasattr(self, key):
            setattr(self, key, value)
        if key in self.data:
            self.data[key] = value
        raise KeyError(f'{key}')

    def prompt(self, model: 'AIModel', **kwarg_overrides) -> 'AIResponse':
        """prompt the ai model with the message

        Args:
            model (AIModel): The model to prmopt

        Returns:
            AIResponse: the response
        """
        return model(self, **kwarg_overrides)

    def reader(self) -> 'Reader':
        """Get the reader for the prompt

        Returns:
            Reader: Returns the base reader
        """
        # TODO: update this to allow it to change
        return NullRead()

    def clone(self) -> typing.Self:
        """Do a shallow copy of the message

        Returns:
            Message: The cloned message
        """
        return self.__class__(
            source=self.source,
            data=self.data
        )

    def aslist(self) -> typing.List['Message']:
        """Get the message as a list

        Returns:
            typing.List[Message]: The message as a list
        """
        return [self]
    
    def render(self) -> str:
        """Render the message

        Returns:
            str: Return the message and the source
        """
        return f'{self.source}: {render(self.data)}'


class TextMessage(Message):
    """A message that contains text. Typically used for LLMs
    """

    def __init__(self, source: str, text: typing.Union[str, 'Cue']):
        """Create a text message with a source

        Args:
            source (str): the source of the message
            text (typing.Union[str, Cue]): the content of the message

        """
        super().__init__(
            source=source,
            data={
                'text': text
            }
        )

    def reader(self) -> 'Reader':
        """The reader to use

        Returns:
            Reader: The reader used by message
        """
        text = self['text']
        if isinstance(text, Cue) and text.out is not None:
            return text.out
        return NullRead(name='')
    
    def render(self) -> str:
        """Render the text message

        Returns:
            str: Return the message and the text for the message
        """
        text = self.data['text']
        return f'{self.source}: {
            text.render() if isinstance(text, Cue) else text
        }'

    def clone(self) -> typing.Self:
        """Do a shallow copy of the message

        Returns:
            Message: The cloned message
        """
        return self.__class__(
            source=self.source,
            text=self.data['text']
        )

    @property
    def text(self) -> str:
        """Get the text for the message

        Returns:
            str: The text for the message
        """
        return self.data['text']


def stream_text(stream: typing.Iterator[typing.Tuple[AIResponse, AIResponse]]) -> typing.Iterator[TextMessage]:
    """Stream the text response from an AI Engine

    Args:
        stream (typing.Iterator[typing.Tuple[AIResponse, AIResponse]]): The stream to execute

    Returns:
        typing.Iterator[TextMessage]: 

    Yields:
        Iterator[typing.Iterator[TextMessage]]: 
    """
    
    for _, a2 in stream:
        yield a2.message


class Dialog(pydantic.BaseModel):
    """A Dialog stores the interactions between the system/user and the assistant
    (i.e. the prompts and the responses)
    """

    messages: typing.List[Message] = pydantic.Field(default_factory=list)

    def __init__(self, messages):
        """Create a dialog

        Args:
            messages: The messages
        """
        super().__init__(messages=messages)

    def __iter__(self) -> typing.Iterator[Message]:
        """Iterate over each message in the dialog

        Yields:
            Iterator[typing.Iterator[Message]]: Each message in the dialog
        """
        for message in self.messages:
            yield message

    def __add__(self, other: 'Dialog') -> 'Dialog':
        """Concatenate two dialogs together

        Args:
            other (Dialog): The other dialog to concatenate

        Returns:
            Dialog: The concatenated dialog
        """
        return Dialog(
            self.messages + other.messages
        )

    def __getitem__(self, idx) -> Message:
        """Retrieve a value from the dialog

        Args:
            idx : The index to add at

        Returns:
            Message: The message in the dialog
        """
        return self.messages[idx]

    def __setitem__(self, idx, message) -> Self:
        """Set idx with a message

        Args:
            idx: The index to set
            message: The message to set

        Returns:
            Dialog: The updated dialog
        """
        self.messages[idx] = message
        return self

    def insert(self, index: int, message: Message):
        """Insert a value into the dialog

        Args:
            index (int): The index to insert at
            message (Message): The message to insert
        """
        self.messages.insert(index, message)

    def pop(self, index: int):
        """Remove a value from the dialog

        Args:
            index (int): The index to pop
        """
        self.messages.pop(index)

    def remove(self, message: Message):
        """Remove a message from the dialog

        Args:
            message (Message): The message to remove
        """
        self.messages.remove(message)

    def append(self, message: Message):
        """Append a message to the end of the dialog

        Args:
            message (Message): The message to add
        """
        self.messages.append(message)

    def add(self, message: Message, ind: typing.Optional[int]=None, replace: bool=False):
        """Add a message to the dialog

        Args:
            message (Message): The message to add
            ind (typing.Optional[int], optional): The index to add. Defaults to None.
            replace (bool, optional): Whether to replace at the index. Defaults to False.

        Raises:
            ValueError: If the index is not correct
        """
        if ind is not None and ind < 0:
            ind = max(len(self.messages) + ind, 0)

        if ind is None or ind == len(self.messages):
            if not replace or ind == len(self.messages):
                self.messages.append(message)
            else:
                self.messages[-1] = message
        elif ind > len(self.messages):
            raise ValueError(
                f'The index {ind} is out of bounds '
                f'for size {len(self.messages)}')
        elif replace:
            self.messages[ind] = message
        else:
            self.messages.insert(ind, message)
        
    def clone(self) -> 'Dialog':
        """Clones the dialog

        Returns:
            Dialog: A dialog cloned with shallow copying of the messages
        """
        return Dialog(
            messages=[message.clone() for message in self.messages]
        )

    def extend(self, dialog: typing.Union['Dialog', typing.List[Message]]):
        """Extend the dialog with another dialog or a list of messages

        Args:
            dialog (typing.Union[&#39;Dialog&#39;, typing.List[Message]]): _description_
        """
        if isinstance(dialog, Dialog):
            dialog = dialog.messages
        
        self.messages.extend(dialog)

    def message(self, source: str, text: typing.Optional[str]=None, _ind: typing.Optional[int]=None, _replace: bool=False, **kwargs):
        """Add a message to the 

        Args:
            source (str): the source of the message
            text (typing.Optional[str], optional): The text message. Defaults to None.
            _ind (typing.Optional[int], optional): The index to set at. Defaults to None.
            _replace (bool, optional): Whether to replace the the text at the index. Defaults to False.

        Raises:
            ValueError: If no message was passed in
        """
        if len(kwargs) == 0 and text is not None:
            message = TextMessage(source, text)
        elif text is not None:
            message = Message(text=text, **kwargs)
        elif text is None:
            message = Message(**kwargs)
        else:
            raise ValueError('No message has been passed. The text and kwargs are empty')

        self.add(message, _ind, _replace)

    def user(self, text: str=None, _ind: int=None, _replace: bool=False, **kwargs):
        """Add a user message

        Args:
            text (str, optional): The text for the message. Defaults to None.
            _ind (int, optional): The index to add to. Defaults to None.
            _replace (bool, optional): Whether to replace at the index. Defaults to False.
        """
        self.message('user', text, _ind, _replace, **kwargs)

    def assistant(self, text: str=None, _ind=None, _replace: bool=False, **kwargs):
        """Add an assistant message

        Args:
            text (str, optional): The text for the message. Defaults to None.
            _ind (int, optional): The index to add to. Defaults to None.
            _replace (bool, optional): Whether to replace at the index. Defaults to False.
        """
        self.message('assistant', text, _ind, _replace, **kwargs)

    def system(self, text: str=None, _ind=None, _replace: bool=False, **kwargs):
        """Add a system message

        Args:
            text (str, optional): The text for the message. Defaults to None.
            _ind (int, optional): The index to add to. Defaults to None.
            _replace (bool, optional): Whether to replace at the index. Defaults to False.
        """
        self.message('system', text, _ind, _replace, **kwargs)

    def reader(self) -> 'Reader':
        """Get the "Reader" for the dialog. By default will use the last one
        that is available.

        Returns:
            Reader: The reader to retrieve
        """
        for r in reversed(self.messages):
            if isinstance(r, Cue):
                if r.reader is not None:
                    return r.reader
        return NullRead(name='')
    
    def exclude(self, *source: str) -> 'Dialog':

        exclude = set(source)
        return Dialog(
            messages=[message for message in self.messages
            if message.source not in exclude]
        )
    
    def include(self, *source: str) -> 'Dialog':
        include = set(source)
        return Dialog(
            messages=[message for message in self.messages
            if message.source in include]
        )

    def render(self) -> str:
        """Render the dialog as a series of turns 
        <source>: <text>

        Returns:
            str: The dialog
        """
        return '\n'.join(
            message.render() for message in self.messages
        )

    def instruct(
        self, instruct: 'Instruct', ai_model: 'AIModel', 
        ind: int=0, replace: bool=True
    ) -> AIResponse:
        """Instruct the AI

        Args:
            instruct (Instruct): The cue to use
            ai_model (AIModel): The AIModel to use
            ind (int, optional): The index to set to. Defaults to 0.
            replace (bool, optional): Whether to replace at the index if already set. Defaults to True.

        Returns:
            AIResponse: The output from the AI
        """
        cue = instruct.i()
        
        self.system(cue, ind, replace)
        response = ai_model.forward(self.messages)
        response = self.process_response(response)
        self.assistant(response.content)
        return response

    def prompt(self, model: 'AIModel', append: bool=True) -> 'AIResponse':
        """Prompt the AI

        Args:
            model (AIModel): The model to usee
            append (bool, optional): Whether to append the output. Defaults to True.

        Returns:
            AIResponse: The response from the AI
        """
        response = model(self)

        if append:
            self.append(response.message)
        return response

    def stream_prompt(self, model: 'AIModel', append: bool=True, **kwarg_override) -> typing.Iterator[typing.Tuple['AIResponse', 'AIResponse']]:
        """Prompt the AI

        Args:
            model (AIModel): The model to usee
            append (bool, optional): Whether to append the output. Defaults to True.

        Returns:
            AIResponse: The response from the AI
        """
        for d, dx in model.stream_forward(self, **kwarg_override):
            yield d, dx

        if append:
            self.append(d.message)
        return d

    def aslist(self) -> typing.List['Message']:
        """Retrieve the message list

        Returns:
            typing.List[Message]: the messages in the dialog
        """
        return self.messages
    
    def __len__(self) -> int:
        """Get the size of the dialog

        Returns:
            int: the number of turns in the dialog
        """
        return len(self.messages)


Data = typing.Union[pydantic.BaseModel, typing.List[pydantic.BaseModel]]


class AIModel(Module, ABC):
    """APIAdapter allows one to adapt various WebAPI or otehr
    API for a consistent interface
    """

    @abstractmethod
    def forward(self, prompt: AIPrompt, **kwarg_override) -> AIResponse:
        """Run a standard query to the API

        Args:
            data : Data to pass to the API

        Returns:
            typing.Dict: The result of the API call
        """
        pass

    @abstractmethod
    def convert(self, message: Message) -> typing.Dict:
        """Convert a message to the format needed for the model

        Args:
            messages (Message): The messages to convert

        Returns:
            typing.List[typing.Dict]: The format to pass to the "model"
        """
        pass

    def convert_messages(self, messages: typing.List[Message]) -> typing.List[typing.Dict]:
        """Convienence method to convert a list of messages to the format needed for the model

        Args:
            messages (typing.List[Message]): The messages to convert

        Returns:
            typing.List[typing.Dict]: The format to pass to the "model"
        """
        return [self.convert(message) for message in messages]

    def stream_forward(self, prompt: AIPrompt, **kwarg_override) -> typing.Iterator[typing.Tuple[AIResponse, AIResponse]]:
        """API that allows for streaming the response

        Args:
            prompt (AIPrompt): Data to pass to the API

        Returns:
            typing.Iterator: Data representing the streamed response
            Uses 'delta' for the difference. Since the default
            behavior doesn't truly stream. This must be overridden 

        Yields:
            typing.Dict: The data
        """
        result = self.forward(prompt, **kwarg_override)
        yield result, result
    
    async def async_forward(
        self, prompt: AIPrompt, **kwarg_override
    ) -> AIResponse:
        """Run this query for asynchronous operations
        The default behavior is simply to call the query

        Args:
            data: Data to pass to the API

        Returns:
            typing.Any: 
        """
        return self.forward(prompt, **kwarg_override)
    
    async def bulk_async_forward(
        self, prompt: typing.List[AIPrompt], **kwarg_override
    ) -> typing.List[AIResponse]:
        """

        Args:
            messages (typing.List[typing.List[Message]]): 

        Returns:
            typing.List[typing.Dict]: 
        """
        tasks = []
        async with asyncio.TaskGroup() as tg:

            for prompt_i in prompt:
                tasks.append(
                    tg.create_task(self.async_forward(prompt_i, **kwarg_override))
                )
        return list(
            task.result() for task in tasks
        )
    
    async def async_stream_forward(self, prompt: AIPrompt, **kwarg_override) -> typing.AsyncIterator[AIResponse]:
        """Run this query for asynchronous streaming operations
        The default behavior is simply to call the query

        Args:
            prompt (AIPrompt): The data to pass to the API

        Yields:
            typing.Dict: The data returned from the API
        """
        result = self.forward(prompt, **kwarg_override)
        yield result

    async def _collect_results(generator, index, results, queue):
        
        async for item in generator:
            results[index] = item
            await queue.put(results[:])  # Put a copy of the current results
        results[index] = None  # Mark this generator as completed

    async def bulk_async_stream_forward(
        self, prompts: typing.List[AIPrompt], **kwarg_override
    ) -> typing.AsyncIterator[typing.List[AIResponse]]:
        """Process multiple 

        Args:
            prompts (AIPrompt): The prompts to process

        Returns:
            typing.List[typing.Dict]: 
        """
        results = [None] * len(prompts)
        queue = asyncio.Queue()

        async with asyncio.TaskGroup() as tg:
            for index, prompt_i in enumerate(prompts):
                tg.create_task(self._collect_results(
                    self.async_stream_forward(prompt_i, **kwarg_override), index, results, queue)
                )

        active_generators = len(prompts)
        while active_generators > 0:
            current_results = await queue.get()
            yield current_results
            active_generators = sum(result is not None for result in current_results)

    def __call__(self, prompt: AIPrompt, **kwarg_override) -> AIResponse:
        """Execute the AIModel

        Args:
            prompt (AIPrompt): The prompt

        Returns:
            AIResponse: Get the response from the AI
        """
        return self.forward(prompt, **kwarg_override)
